VERSION = "1.0.95"
# noinspection PyBroadException
try:
    from subprocess import check_output, DEVNULL
    GIT_VERSION = check_output(["git", "describe", "--tags", "--always"], stderr=DEVNULL).decode().strip().lstrip("v")
except:
    GIT_VERSION = VERSION



from ucca import layer0, layer1
from ucca.layer0 import NodeTags as L0Tags
from ucca.layer1 import EdgeTags as ETags, NodeTags as L1Tags

LINKAGE = (ETags.LinkArgument, ETags.LinkRelation)


def validate(passage):
    for node in passage.layer(layer0.LAYER_ID).all:
        yield from NodeValidator(node).validate_terminal()
    heads = list(passage.layer(layer1.LAYER_ID).heads)
    linkage = False
    for node in heads:
        if node.tag == L1Tags.Linkage:
            linkage = True
        yield from NodeValidator(node).validate_top_level()
    stack = [heads]
    visited = set()
    path = []
    path_set = set(path)
    while stack:
        for node in stack[-1]:
            if node in path_set:
                yield "Detected cycle (%s)" % "->".join(n.ID for n in path)
            elif node not in visited:
                visited.add(node)
                path.append(node)
                path_set.add(node)
                stack.append(node.children)
                yield from NodeValidator(node).validate_non_terminal(linkage)
                break
        else:
            if path:
                path_set.remove(path.pop())
            stack.pop()


class NodeValidator:
    def __init__(self, node):
        self.node = node
        self.incoming = tag_to_edge(node.incoming)
        self.outgoing = tag_to_edge(node)
        self.incoming_tags = set(self.incoming)
        self.outgoing_tags = set(self.outgoing)

    def validate_terminal(self):
        if not self.incoming:
            yield "Orphan %s terminal (%s) '%s'" % (self.node.tag, self.node.ID, self.node)
        elif len(self.node.incoming) > 1:
            yield "Reentrant %s terminal (%s) '%s'" % (self.node.tag, join(self.node.incoming), self.node)

    def validate_top_level(self):
        if self.node.ID != "1.1" and self.node.tag != L1Tags.Linkage:
            yield "Extra root (%s)" % self.node.ID
        terminals = [n for n in self.node.children if n.layer.ID == layer0.LAYER_ID]
        if terminals:
            yield "Terminal children (%s) of root (%s)" % (join(terminals), self.node)
        s = self.outgoing_tags.difference((ETags.ParallelScene, ETags.Linker, ETags.Function, ETags.Ground,
                                           ETags.Punctuation, ETags.LinkRelation, ETags.LinkArgument))
        if s:
            yield "Top-level node (%s) with %s edge" % (self.node.ID, join(s))

    def validate_non_terminal(self, linkage=False):
        if self.node.tag == L1Tags.Linkage:
            yield from self.validate_linkage()
        elif self.node.tag == L1Tags.Foundational:
            yield from self.validate_foundational()
        primary_incoming = [e for e in self.node.incoming if not e.attrib.get("remote") and e.tag not in LINKAGE]
        if len(primary_incoming) > 1:
            yield "Multiple incoming non-remote (%s)" % join(primary_incoming)
        remote_incoming = [e for e in self.node.incoming if e.attrib.get("remote")]
        if remote_incoming and not primary_incoming:
            yield "Node (%s) with remote parents but no primary parents" % self.node.ID
        for edge in self.node:
            if (edge.tag == ETags.Punctuation) != (edge.child.tag == L1Tags.Punctuation):
                yield "%s edge (%s) with %s child" % (edge.tag, edge, edge.child.tag)
            if (self.node.tag == L1Tags.Punctuation) != (edge.child.tag == L0Tags.Punct):
                yield "%s node (%s) with %s child (%s)" % (self.node.tag, self.node.ID, edge.child.tag, edge.child.ID)
        if self.node.attrib.get("implicit") and self.node.outgoing:
            yield "Implicit node (%s) with outgoing edges (%s)" % (self.node.ID, join(self.node))
        for tag in (ETags.Function, ETags.Ground, ETags.ParallelScene, ETags.Linker, ETags.LinkRelation,
                    ETags.Connector, ETags.Punctuation, ETags.Terminal):
            s = self.incoming.get(tag, ())
            if len(s) > 1:
                yield "Multiple incoming %s edges (%s)" % (tag, join(s))
        for tag in (ETags.LinkRelation, ETags.Process, ETags.State):
            s = self.outgoing.get(tag, ())
            if len(s) > 1:
                yield "Multiple outgoing %s edges (%s)" % (tag, join(s))
        if ETags.Function in self.incoming:
            s = self.outgoing_tags.difference((ETags.Terminal, ETags.Punctuation))
            if s:
                yield "%s node (%s) with outgoing %s edge" % (ETags.Function, self.node.ID, join(s))
        if ETags.Linker in self.incoming_tags:
            s = self.incoming_tags.difference((ETags.Linker, ETags.LinkRelation))
            if s:
                yield "%s node (%s) with incoming %s edge" % (ETags.Linker, self.node.ID, join(s))
            if linkage and ETags.LinkRelation not in self.incoming_tags:
                yield "%s node (%s) with no incoming %s" % (ETags.Linker, self.node.ID, ETags.LinkRelation)

    def validate_linkage(self):
        if self.node.incoming:
            yield "Non-root %s node (%s)" % (self.node.tag, self.node)
        s = self.outgoing_tags.difference(LINKAGE)
        if s:
            yield "%s node (%s) with %s children" % (self.node.tag, self.node, join(s))
        if ETags.LinkRelation not in self.outgoing:
            yield "%s node without %s child" % (self.node.tag, ETags.LinkRelation)

    def validate_foundational(self):
        if self.node.participants and not self.node.is_scene():
            yield "Node (%s) with participants but without main relation" % self.node.ID
        if self.node.process and self.node.state:
            yield "Node (%s) with both process (%s) and state (%s)" % (self.node.ID, self.node.process, self.node.state)
        if self.node.parallel_scenes:
            s = self.outgoing_tags.difference((ETags.ParallelScene, ETags.Punctuation, ETags.Linker,
                                               ETags.Ground, ETags.Relator, ETags.Function))
            if s:
                yield "Node (%s) with parallel scenes has %s edge" % (self.node.ID, join(s))
        s = self.outgoing_tags.intersection(LINKAGE)
        if s:
            yield "Non-linkage node (%s) with %s edges" % (self.node, join(s))


def tag_to_edge(edges):
    d = {}
    for edge in edges:
        d.setdefault(edge.tag, []).append(edge)
    return d


def join(items):
    return ", ".join(map(str, items))



#!/usr/bin/env python3

import argparse

from ucca.ioutil import write_passage, get_passages_with_progress_bar
from ucca.textutil import annotate_all, is_annotated

desc = """Read UCCA standard format in XML or binary pickle, and write back with POS tags and dependency parse."""


def main(args):
    for passage in annotate_all(get_passages_with_progress_bar(args.filenames, desc="Annotating"),
                                replace=True, as_array=args.as_array, verbose=args.verbose):
        assert is_annotated(passage, args.as_array), "Passage %s is not annotated" % passage.ID
        write_passage(passage, outdir=args.out_dir, verbose=args.verbose)


if __name__ == '__main__':
    argparser = argparse.ArgumentParser(description=desc)
    argparser.add_argument("filenames", nargs="+", help="passage file names to annotate")
    argparser.add_argument("-o", "--out-dir", default=".", help="directory to write annotated files to")
    argparser.add_argument("-a", "--as-array", action="store_true", help="save annotations as array in passage level")
    argparser.add_argument("-v", "--verbose", action="store_true", help="print tagged text for each passage")
    main(argparser.parse_args())



import pytest

from ucca import core, layer0, layer1
from ucca.evaluation import evaluate, LABELED, UNLABELED
from .conftest import PASSAGES

"""Tests the evaluation module functions and classes."""


def passage1():
    p = core.Passage("1")
    l0 = layer0.Layer0(p)
    l1 = layer1.Layer1(p)
    # 20 terminals (1-20), #10 and #20 are punctuation
    terms = [l0.add_terminal(text=str(i), punct=(i % 10 == 0)) for i in range(1, 21)]

    # Linker #1 with terminal 1
    link1 = l1.add_fnode(None, layer1.EdgeTags.Linker)
    link1.add(layer1.EdgeTags.Terminal, terms[0])

    # Scene #1: [[2 3 4 5 P] [6 7 8 9 A] [10 U] H]
    ps1 = l1.add_fnode(None, layer1.EdgeTags.ParallelScene)
    p1 = l1.add_fnode(ps1, layer1.EdgeTags.Process)
    a1 = l1.add_fnode(ps1, layer1.EdgeTags.Participant)
    p1.add(layer1.EdgeTags.Terminal, terms[1])
    p1.add(layer1.EdgeTags.Terminal, terms[2])
    p1.add(layer1.EdgeTags.Terminal, terms[3])
    p1.add(layer1.EdgeTags.Terminal, terms[4])
    a1.add(layer1.EdgeTags.Terminal, terms[5])
    a1.add(layer1.EdgeTags.Terminal, terms[6])
    a1.add(layer1.EdgeTags.Terminal, terms[7])
    a1.add(layer1.EdgeTags.Terminal, terms[8])
    l1.add_punct(ps1, terms[9])

    # Scene #23: [[11 12 13 14 15 H] [16 L] [17 18 19 H] H]
    # Scene #2: [[11 12 13 14 P] [15 D]]
    ps23 = l1.add_fnode(None, layer1.EdgeTags.ParallelScene)
    ps2 = l1.add_fnode(ps23, layer1.EdgeTags.ParallelScene)
    a2 = l1.add_fnode(ps2, layer1.EdgeTags.Participant)
    a2.add(layer1.EdgeTags.Terminal, terms[10])
    a2.add(layer1.EdgeTags.Terminal, terms[11])
    a2.add(layer1.EdgeTags.Terminal, terms[12])
    a2.add(layer1.EdgeTags.Terminal, terms[13])
    d2 = l1.add_fnode(ps2, layer1.EdgeTags.Adverbial)
    d2.add(layer1.EdgeTags.Terminal, terms[14])

    # Linker #2: [16 L]
    link2 = l1.add_fnode(ps23, layer1.EdgeTags.Linker)
    link2.add(layer1.EdgeTags.Terminal, terms[15])

    # Scene #3: [[16 17 S] [18 A] (implicit participant) H]
    ps3 = l1.add_fnode(ps23, layer1.EdgeTags.ParallelScene)
    p3 = l1.add_fnode(ps3, layer1.EdgeTags.State)
    p3.add(layer1.EdgeTags.Terminal, terms[16])
    p3.add(layer1.EdgeTags.Terminal, terms[17])
    a3 = l1.add_fnode(ps3, layer1.EdgeTags.Participant)
    a3.add(layer1.EdgeTags.Terminal, terms[18])
    l1.add_fnode(ps3, layer1.EdgeTags.Participant, implicit=True)

    # Punctuation #20 - not under a scene
    l1.add_punct(None, terms[19])

    # adding remote argument to scene #1, remote process to scene #2
    # creating linkages L1->H1, H2<-L2->H3
    l1.add_remote(ps1, layer1.EdgeTags.Participant, d2)
    l1.add_remote(ps2, layer1.EdgeTags.Process, p1)
    l1.add_linkage(link1, ps1)
    l1.add_linkage(link2, ps2, ps3)

    return p


def passage2():
    p = core.Passage("2")
    l0 = layer0.Layer0(p)
    l1 = layer1.Layer1(p)
    # 20 terminals (1-20), #10 and #20 are punctuation
    terms = [l0.add_terminal(text=str(i), punct=(i % 10 == 0)) for i in range(1, 21)]

    # Linker #1 with terminal 1
    link1 = l1.add_fnode(None, layer1.EdgeTags.Linker)
    link1.add(layer1.EdgeTags.Terminal, terms[0])

    # Scene #1: [[2 3 4 5 P] [6 7 8 9 A] [10 U] H]
    ps1 = l1.add_fnode(None, layer1.EdgeTags.ParallelScene)
    p1 = l1.add_fnode(ps1, layer1.EdgeTags.Process)
    a1 = l1.add_fnode(ps1, layer1.EdgeTags.Participant)
    p1.add(layer1.EdgeTags.Terminal, terms[1])
    p1.add(layer1.EdgeTags.Terminal, terms[2])
    p1.add(layer1.EdgeTags.Terminal, terms[3])
    p1.add(layer1.EdgeTags.Terminal, terms[4])
    a1.add(layer1.EdgeTags.Terminal, terms[5])
    a1.add(layer1.EdgeTags.Terminal, terms[6])
    a1.add(layer1.EdgeTags.Terminal, terms[7])
    a1.add(layer1.EdgeTags.Terminal, terms[8])
    l1.add_punct(ps1, terms[9])

    # Scene #23: [[11 12 13 14 15 H] [16 L] [17 18 19 H] H]
    # Scene #2: [[11 12 13 14 P] [15 D]]
    ps23 = l1.add_fnode(None, layer1.EdgeTags.ParallelScene)
    ps2 = l1.add_fnode(ps23, layer1.EdgeTags.ParallelScene)
    a2 = l1.add_fnode(ps2, layer1.EdgeTags.ParallelScene)
    a2.add(layer1.EdgeTags.Terminal, terms[10])
    a2.add(layer1.EdgeTags.Terminal, terms[11])
    a2.add(layer1.EdgeTags.Terminal, terms[12])
    a2.add(layer1.EdgeTags.Terminal, terms[13])
    d2 = l1.add_fnode(ps1, layer1.EdgeTags.Elaborator)
    d2.add(layer1.EdgeTags.Terminal, terms[14])

    # Linker #2: [16 L]
    link2 = l1.add_fnode(ps23, layer1.EdgeTags.Linker)
    link2.add(layer1.EdgeTags.Terminal, terms[15])

    # Scene #3: [[16 17 S] [18 A] (implicit participant) H]
    ps3 = l1.add_fnode(ps23, layer1.EdgeTags.ParallelScene)
    p3 = l1.add_fnode(ps3, layer1.EdgeTags.Process)
    p3.add(layer1.EdgeTags.Terminal, terms[16])
    p3.add(layer1.EdgeTags.Terminal, terms[17])
    a3 = l1.add_fnode(ps3, layer1.EdgeTags.Participant)
    a3.add(layer1.EdgeTags.Terminal, terms[18])
    l1.add_fnode(ps3, layer1.EdgeTags.Participant, implicit=True)

    # Punctuation #20 - not under a scene
    l1.add_punct(None, terms[19])

    # adding remote argument to scene #1, remote process to scene #2
    # creating linkages L1->H1, H2<-L2->H3
    l1.add_remote(ps1, layer1.EdgeTags.Participant, d2)
    l1.add_remote(ps1, layer1.EdgeTags.Participant, a3)
    l1.add_remote(ps2, layer1.EdgeTags.State, p1)
    l1.add_linkage(link1, ps1)
    l1.add_linkage(link2, ps2, ps3)

    return p


def check_primary_remote(scores, primary_labeled_f1, remote_labeled_f1, primary_unlabeled_f1, remote_unlabeled_f1):
    assert scores[LABELED]["primary"].f1 == primary_labeled_f1
    assert scores[LABELED]["remote"].f1 == remote_labeled_f1
    assert scores[UNLABELED]["primary"].f1 == primary_unlabeled_f1
    assert scores[UNLABELED]["remote"].f1 == remote_unlabeled_f1


@pytest.mark.parametrize("create", PASSAGES + (passage1, passage2))
@pytest.mark.parametrize("units", (True, False), ids=("units", ""))
@pytest.mark.parametrize("errors", (True, False), ids=("errors", ""))
@pytest.mark.parametrize("normalize", (True, False), ids=("normalize", ""))
def test_evaluate_self(create, units, errors, normalize):
    p = create()
    scores = evaluate(p, p, units=units, errors=errors, normalize=normalize)
    assert scores.average_f1() == 1.0
    for eval_type, results in scores.evaluators.items():
        for construction, stats in results.results.items():
            assert stats.f1 == 1.0, (eval_type, construction)
            assert stats.p == 1.0, (eval_type, construction)
            assert stats.r == 1.0, (eval_type, construction)
    check_primary_remote(scores, 1.0, 1.0, 1.0, 1.0)


@pytest.mark.parametrize("create1, create2, primary_labeled_f1, remote_labeled_f1,"
                         "primary_unlabeled_f1, remote_unlabeled_f1", (
                                 (passage1, passage2, 0.5, 0.4, 0.75, 0.8),
                         ))
@pytest.mark.parametrize("units", (True, False), ids=("units", ""))
@pytest.mark.parametrize("errors", (True, False), ids=("errors", ""))
def test_evaluate(create1, create2, primary_labeled_f1, remote_labeled_f1, primary_unlabeled_f1, remote_unlabeled_f1,
                  units, errors):
    scores = evaluate(create1(), create2(), units=units, errors=errors)
    check_primary_remote(scores, primary_labeled_f1, remote_labeled_f1, primary_unlabeled_f1, remote_unlabeled_f1)



import pytest

from ucca.visualization import draw, tikz
from .conftest import PASSAGES

"""Tests the visualization module functions and classes."""


@pytest.mark.parametrize("create", PASSAGES)
def test_draw(create):
    import matplotlib
    matplotlib.use('Agg')
    draw(create())


@pytest.mark.parametrize("create", PASSAGES)
def test_tikz(create):
    tikz(create())



import pytest

from ucca.validation import validate
from .conftest import loaded, loaded_valid, multi_sent, crossing, discontiguous, l1_passage, empty

"""Tests the validation module functions and classes."""


@pytest.mark.parametrize("create, valid", (
        (loaded, False),
        (loaded_valid, True),
        (multi_sent, True),
        (crossing, True),
        (discontiguous, True),
        (l1_passage, True),
        (empty, True),
))
def test_evaluate_self(create, valid):
    p = create()
    errors = list(validate(p))
    if valid:
        assert not errors, p
    else:
        assert errors, p



#!/usr/bin/env python3
"""
The evaluation software for UCCA layer 1.
"""

from argparse import ArgumentParser

from scripts import ucca_db
from ucca import convert, constructions
from ucca.evaluation import evaluate


def main(args):
    keys = [args.guessed, args.ref]
    xmls = ucca_db.get_by_xids(args.db_filename, args.host, keys) if args.from_xids else \
        ucca_db.get_xml_trees(args.db_filename, args.host, args.pid, keys)
    guessed, ref = [convert.from_site(x) for x in xmls]
    if args.units or args.fscore or args.errors:
        evaluate(guessed, ref, units=args.units, fscore=args.fscore, errors=args.errors,
                 constructions=args.constructions, verbose=True)


if __name__ == '__main__':
    argparser = ArgumentParser(description="Evaluate passages on UCCA DB")
    argparser.add_argument("--db", "-d", required=True, dest="db_filename", help="the db file name")
    argparser.add_argument("--host", "--hst", help="the host name")
    group = argparser.add_mutually_exclusive_group()
    group.add_argument("-p", "--pid", type=int, help="the passage ID")
    group.add_argument("-x", "--from_xids", action="store_true",
                       help="interpret the ref and the guessed parameters as Xids in the db")
    argparser.add_argument("--guessed", "-g", required=True,
                           help="if a db is defined - the username for the guessed annotation; "
                                "else - the xml file name for the guessed annotation")
    argparser.add_argument("-r", "--ref", required=True,
                           help="if a db is defined - the username for the reference annotation; "
                                "else - the xml file name for the reference annotation")
    argparser.add_argument("-u", "--units", action="store_true",
                           help="the units the annotations have in common, and those each has separately")
    argparser.add_argument("-f", "--fscore", action="store_true",
                           help="outputs the traditional P,R,F instead of the scene structure evaluation")
    argparser.add_argument("-e", "--errors", action="store_true",
                           help="prints the error distribution according to its frequency")
    constructions.add_argument(argparser)
    main(argparser.parse_args())



from time import sleep

import json
import logging
import os
import requests

DEFAULT_SERVER = "http://ucca-demo.cs.huji.ac.il"
API_PREFIX = "/api/v1/"
SERVER_ADDRESS_ENV_VAR = "UCCA_APP_SERVER_ADDRESS"
AUTH_TOKEN_ENV_VAR = "UCCA_APP_AUTH_TOKEN"
EMAIL_ENV_VAR = "UCCA_APP_EMAIL"
PASSWORD_ENV_VAR = "UCCA_APP_PASSWORD"
PROJECT_ID_ENV_VAR = "UCCA_APP_PROJECT_ID"
SOURCE_ID_ENV_VAR = "UCCA_APP_SOURCE_ID"
USER_ID_ENV_VAR = "UCCA_APP_USER_ID"
MAX_RETRIES = 3
RETRY_WAIT_DURATION = 60


class ServerAccessor:
    def __init__(self, server_address, email, password, auth_token, project_id, source_id, verbose, **kwargs):
        if verbose:
            logging.basicConfig(level=logging.DEBUG)
        server_address = server_address or os.environ.get(SERVER_ADDRESS_ENV_VAR, DEFAULT_SERVER)
        self.prefix = server_address + API_PREFIX
        self.headers = {}  # Needed for self.request (login)
        token = auth_token or os.environ.get(AUTH_TOKEN_ENV_VAR) or self.login(
            email or os.environ[EMAIL_ENV_VAR], password or os.environ[PASSWORD_ENV_VAR])["token"]
        self.headers["Authorization"] = "Token " + token
        self.source = self.get_source(source_id or int(os.environ[SOURCE_ID_ENV_VAR]))
        self.project = self.get_project(project_id or int(os.environ[PROJECT_ID_ENV_VAR]))
        self.layer = self.get_layer(self.project["layer"]["id"])
        self.user = None

    def set_user(self, user_id=None):
        self.user = dict(id=user_id or int(os.environ[USER_ID_ENV_VAR]))

    @staticmethod
    def add_arguments(argparser):
        argparser.add_argument("--server-address", help="UCCA-App server, otherwise set by " + SERVER_ADDRESS_ENV_VAR)
        argparser.add_argument("--email", help="UCCA-App email, otherwise set by " + EMAIL_ENV_VAR)
        argparser.add_argument("--password", help="UCCA-App password, otherwise set by " + PASSWORD_ENV_VAR)
        argparser.add_argument("--auth-token", help="authorization token (required only if email or password missing), "
                                                    "otherwise set by " + AUTH_TOKEN_ENV_VAR)
        argparser.add_argument("--project-id", type=int, help="project id, otherwise set by " + PROJECT_ID_ENV_VAR)
        argparser.add_argument("--source-id", type=int, help="source id, otherwise set by " + SOURCE_ID_ENV_VAR)
        argparser.add_argument("-v", "--verbose", action="store_true", help="detailed output")

    @staticmethod
    def add_user_id_argument(argparser):
        argparser.add_argument("--user-id", type=int, help="user id, otherwise set by " + USER_ID_ENV_VAR)

    def request(self, method, url_suffix, **kwargs):
        response = None
        for _ in range(MAX_RETRIES):
            response = requests.request(method, self.prefix + str(url_suffix), headers=self.headers, **kwargs)
            if response.status_code != 500:
                break
            sleep(RETRY_WAIT_DURATION)
        response.raise_for_status()
        return response

    def login(self, email, password):
        return self.request("post", "login", json=dict(email=email, password=password)).json()

    def get_source(self, source_id):
        logging.debug("Getting source %d" % source_id)
        source_out = self.request("get", "sources/%d/" % source_id).json()
        logging.debug("Got source: " + json.dumps(source_out))
        return source_out

    def get_project(self, project_id):
        logging.debug("Getting project %d" % project_id)
        project_out = self.request("get", "projects/%d/" % project_id).json()
        logging.debug("Got project: " + json.dumps(project_out))
        return project_out

    def get_layer(self, layer_id):
        logging.debug("Getting layer %d" % layer_id)
        layer_out = self.request("get", "layers/%d/" % layer_id).json()
        logging.debug("Got layer: " + json.dumps(layer_out))
        return layer_out

    def get_user(self, user_id):
        logging.debug("Getting user "+user_id)
        user_out = self.request("get", "users/%s/" % user_id).json()
        logging.debug("Got user: " + json.dumps(user_out))
        return user_out

    def get_task(self, task_id):
        logging.debug("Getting task " + str(task_id))
        task_out = self.request("get", "tasks/" + str(task_id)).json()
        logging.debug("Got task: " + json.dumps(task_out))
        return task_out

    def get_user_task(self, task_id):
        logging.debug("Getting user task " + str(task_id))
        task_out = self.request("get", "user_tasks/" + str(task_id)).json()
        logging.debug("Got user task: " + json.dumps(task_out))
        return task_out

    def get_passage(self, passage_id):
        logging.debug("Getting passage " + str(passage_id))
        passage_out = self.request("get", "passages/" + str(passage_id)).json()
        logging.debug("Got passage: " + json.dumps(passage_out))
        return passage_out

    def create_passage(self, **kwargs):
        logging.debug("Creating passage: " + json.dumps(kwargs))
        passage_out = self.request("post", "passages/", json=kwargs).json()
        logging.debug("Created passage: " + json.dumps(passage_out))
        return passage_out

    def create_tokenization_task(self, **kwargs):
        logging.debug("Creating tokenization task: " + json.dumps(kwargs))
        tok_task_out = self.request("post", "tasks/", json=kwargs).json()
        logging.debug("Created tokenization task: " + json.dumps(tok_task_out))
        return tok_task_out

    def submit_tokenization_task(self, **kwargs):
        logging.debug("Submitting tokenization task: " + json.dumps(kwargs))
        self.request("put", "user_tasks/%d/draft" % kwargs["id"], json=kwargs)
        tok_user_task_out = self.request("put", "user_tasks/%d/submit" % kwargs["id"]).json()
        logging.debug("Submitted tokenization task: " + json.dumps(tok_user_task_out))
        return tok_user_task_out

    def create_annotation_task(self, **kwargs):
        logging.debug("Creating annotation task: " + json.dumps(kwargs))
        ann_task_out = self.request("post", "tasks/", json=kwargs).json()
        logging.debug("Created annotation task: " + json.dumps(ann_task_out))
        return ann_task_out

    def submit_annotation_task(self, **kwargs):
        logging.debug("Submitting annotation task: " + json.dumps(kwargs))
        self.request("put", "user_tasks/%d/draft" % kwargs["id"], json=kwargs)
        ann_user_task_out = self.request("put", "user_tasks/%d/submit" % kwargs["id"]).json()
        logging.debug("Submitted annotation task: " + json.dumps(ann_user_task_out))
        return ann_user_task_out



import os
import pytest
import random
from glob import glob

from ucca import layer0, layer1, convert, ioutil, diffutil
from ucca.convert import FROM_FORMAT
from .conftest import loaded, multi_sent, discontiguous, l1_passage

"""Tests the ioutil module functions and classes."""


def test_split2sentences():
    """Tests splitting a passage by sentence ends.
    """
    p = multi_sent()
    split = convert.split2sentences(p)
    assert len(split) == 3
    terms = [[t.text for t in s.layer(layer0.LAYER_ID).all] for s in split]
    assert terms[0] == ["1", "2", "3", "."]
    assert terms[1] == ["5", "6", "."]
    assert terms[2] == ["8", ".", "10", "."]
    assert all(t.paragraph == 1 for s in split for t in s.layer(layer0.LAYER_ID).all)
    top_scenes = [s.layer(layer1.LAYER_ID).top_scenes for s in split]
    for t in top_scenes:
        assert len(t) == 1
        assert t[0].incoming[0].tag == layer1.EdgeTags.ParallelScene


def test_split2paragraphs():
    """Tests splitting a passage by paragraph ends.
    """
    p = multi_sent()
    split = convert.split2paragraphs(p)
    assert len(split) == 2
    terms = [[t.text for t in s.layer(layer0.LAYER_ID).all] for s in split]
    assert terms[0] == ["1", "2", "3", ".", "5", "6", "."]
    assert terms[1] == ["8", ".", "10", "."]
    assert all(t.paragraph == 1 for s in split for t in s.layer(layer0.LAYER_ID).all)
    top_scenes = [s.layer(layer1.LAYER_ID).top_scenes for s in split]
    assert len(top_scenes[0]) == 2
    assert len(top_scenes[1]) == 1
    for t in top_scenes:
        for n in t:
            assert n.incoming[0].tag == layer1.EdgeTags.ParallelScene


@pytest.mark.parametrize("create", (loaded, multi_sent, discontiguous, l1_passage))
def test_split_join_sentences(create):
    p = create()
    split = convert.split2sentences(p, remarks=True)
    copy = convert.join_passages(split)
    diffutil.diff_passages(p, copy)
    assert p.equals(copy)


@pytest.mark.parametrize("create", (loaded, multi_sent, discontiguous, l1_passage))
def test_split_join_paragraphs(create):
    p = create()
    split = convert.split2paragraphs(p, remarks=True)
    copy = convert.join_passages(split)
    diffutil.diff_passages(p, copy)
    assert p.equals(copy)


FORMATS = ("xml", "conll", "export", "sdp")


def _test_passages(passages):
    for passage in passages:
        assert passage.layer(layer0.LAYER_ID).all, "No terminals in passage " + passage.ID
        assert len(passage.layer(layer1.LAYER_ID).all), "No non-terminals but the root in passage " + passage.ID


@pytest.mark.parametrize("suffix", FORMATS)
def test_load_passage(suffix):
    _test_passages(ioutil.read_files_and_dirs(glob(os.path.join("test_files", "standard3." + suffix)),
                                              converters=FROM_FORMAT))


def test_load_multiple_passages():
    """Test lazy-loading passages"""
    files = ["test_files/standard3.%s" % s for s in FORMATS]
    passages = ioutil.read_files_and_dirs(files, converters=FROM_FORMAT)
    assert len(files) == len(list(passages)), "Should load one passage per file"
    assert len(files) == len(passages)
    _test_passages(passages)


def test_shuffle_passages():
    """Test lazy-loading passages and shuffling them"""
    files = ["test_files/standard3.%s" % s for s in FORMATS]
    passages = ioutil.read_files_and_dirs(files, converters=FROM_FORMAT)
    print("Passages:\n" + "\n".join(str(p.layer(layer1.LAYER_ID).heads[0]) for p in passages))
    random.shuffle(passages)
    print("Shuffled passages:\n" + "\n".join(str(p.layer(layer1.LAYER_ID).heads[0]) for p in passages))
    assert len(files) == len(passages)
    _test_passages(passages)
